require 'range_utils'

module RangeHeaderParser
  RANGE_RE = /^(\d+)-(\d+)$/.freeze
  RANGE_FROM = /^(\d+)-$/.freeze
  RANGE_LAST_N =/^-(\d+)$/.freeze
  BYTES_RE = /^bytes=(.*)/.freeze
  
  InvalidHeader = Class.new(RuntimeError)
  InvalidRange = Class.new(InvalidHeader)

  # Parses a given string from Range: header,
  # and returns an Array of Range objects.
  #
  # Overlapping and adjacent ranges will be spliced.
  def parse(header_str, resource_size)
    ranges = ranges_from_header_str(header_str, resource_size)
    RangeUtils.splice(ranges)
  end
  
  extend self
  
  private
  
  # Ported from WEBrick::HTTPUtils.parse_range_header, but with extra juice and
  # honoring the actual size of the resource.
  # https://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.35.1
  # If the last-byte-pos value is absent, or if the value is greater than or equal
  # to the current length of the entity-body, last-byte-pos is taken to be equal
  # to one less than the current length of the entity- body in bytes.
  def ranges_from_header_str(ranges_specifier, size)
    raise InvalidHeader, ranges_specifier unless BYTES_RE =~ ranges_specifier
    byte_range_set = split_header_value($1)
    byte_range_set.map do |range_spec|
      case range_spec
      when RANGE_RE # From byte M to byte N
        starts_at = $1.to_i
        ends_at = $2.to_i
        ends_at = (size - 1) if ends_at >= size
        raise InvalidRange, "Invalid range spec  #{range_spec} (starts too far in the resource)" if starts_at >= size
        raise InvalidRange, "Invalid range spec #{range_spec} (start and end are flipped)" if ends_at < starts_at
        starts_at..ends_at
      when RANGE_FROM # From byte N to the end
        starts_at = $1.to_i
        raise InvalidRange, "Invalid range spec (starts too far in the resource) #{range_spec}" if starts_at >= size
        starts_at..(size - 1)
      when RANGE_LAST_N # The last N bytes
        last_n = $1.to_i
        raise InvalidRange, "Invalid range spec #{range_spec} (resource too small)" if last_n > size
        
        starts_at = size - last_n
        ends_at = size - 1
        raise InvalidRange, "Invalid range spec #{range_spec} (flipped negative)" if starts_at >= size
        starts_at..ends_at
      else
        raise InvalidRange, "Invalid range spec #{range_spec} (unparseable)"
      end
    end
  end
  
  # Ported from webrick
  def split_header_value(str)
    str.scan(%r'\G((?:"(?:\.|[^"])+?"|[^",]+)+)
                  (?:,\s*|\Z)'xn).flatten
  end
end
